from enum import Enum
from typing import Type

from django.core import checks

from rest_registration.utils.checks import CheckCode, CheckMessage


class _BaseCheckCodeMixin(CheckCode):  # noqa: E501 pylint: disable=abstract-method

    def get_app_name(self) -> str:
        from rest_registration.apps import RestRegistrationConfig  # noqa: E501 pylint: disable=import-outside-toplevel, cyclic-import
        return RestRegistrationConfig.name


class ErrorCode(_BaseCheckCodeMixin, Enum):
    NO_RESET_PASSWORD_VER_URL = 1
    NO_REGISTER_VER_URL = 2
    NO_REGISTER_EMAIL_VER_URL = 3
    NO_VER_FROM_EMAIL = 4
    NO_TOKEN_AUTH_CONFIG = 5
    NO_TOKEN_AUTH_INSTALLED = 6
    INVALID_EMAIL_TEMPLATE_CONFIG = 7
    NO_AUTH_INSTALLED = 8
    DRF_INCOMPATIBLE_DJANGO_AUTH_BACKEND = 9
    LOGIN_FIELDS_NOT_UNIQUE = 10
    INVALID_AUTH_TOKEN_MANAGER_CLASS = 11
    INVALID_REGISTER_EMAIL_SERIALIZER_CLASS = 12
    NON_UNIQUE_FIELD_USED_AS_UNIQUE = 13

    def get_code_id(self) -> str:
        return 'E{self.value:03d}'.format(self=self)

    def get_check_message_class(self) -> Type[CheckMessage]:
        return checks.Error

    def __str__(self) -> str:
        return self.get_code_id()


class WarningCode(_BaseCheckCodeMixin, Enum):
    REGISTER_VERIFICATION_MULTIPLE_AUTO_LOGIN = 1
    DEPRECATION = 2

    def get_code_id(self) -> str:
        return 'W{self.value:03d}'.format(self=self)

    def get_check_message_class(self) -> Type[CheckMessage]:
        return checks.Warning

    def __str__(self) -> str:
        return self.get_code_id()



from django.shortcuts import redirect
from django.views.decorators.http import require_http_methods
from rest_framework.exceptions import APIException

from rest_registration.api.views.register import process_verify_registration_data
from rest_registration.api.views.register_email import process_verify_email_data
from rest_registration.api.views.reset_password import process_reset_password_data
from rest_registration.contrib.verification_redirects.settings import (
    verification_redirects_settings
)


@require_http_methods(['GET'])
def verify_registration(request):
    return _generic_redirect_view(
        request, process_verify_registration_data,
        ['user_id', 'signature', 'timestamp'],
        verification_redirects_settings.VERIFY_REGISTRATION_SUCCESS_URL,
        verification_redirects_settings.VERIFY_REGISTRATION_FAILURE_URL)


@require_http_methods(['GET'])
def verify_email(request):
    return _generic_redirect_view(
        request, process_verify_email_data,
        ['user_id', 'email', 'signature', 'timestamp'],
        verification_redirects_settings.VERIFY_EMAIL_SUCCESS_URL,
        verification_redirects_settings.VERIFY_EMAIL_FAILURE_URL)


@require_http_methods(['POST'])
def reset_password(request):
    return _generic_redirect_view(
        request, process_reset_password_data,
        ['user_id', 'signature', 'timestamp', 'password'],
        verification_redirects_settings.RESET_PASSWORD_SUCCESS_URL,
        verification_redirects_settings.RESET_PASSWORD_FAILURE_URL,
        use_post_method=True)


def _generic_redirect_view(
        request, data_processor, data_keys, success_url, failure_url,
        use_post_method=False):
    query_dict = request.POST if use_post_method else request.GET
    data = {}
    for key in data_keys:
        data[key] = query_dict.get(key)
    try:
        request = data_processor(data)
        return redirect(success_url)
    except APIException:
        return redirect(failure_url)



from django.contrib.auth import get_user_model
from rest_framework import serializers

from rest_registration.settings import registration_settings
from rest_registration.utils.users import get_user_public_field_names
from rest_registration.utils.validation import (
    run_validators,
    validate_user_password,
    validate_user_password_confirm
)


class MetaObj:
    pass


class PasswordConfirmSerializerMixin(serializers.Serializer):

    def has_password_confirm_field(self) -> bool:
        raise NotImplementedError()

    def get_fields(self):
        fields = super().get_fields()
        if self.has_password_confirm_field():
            fields['password_confirm'] = serializers.CharField(write_only=True)
        return fields


class DefaultLoginSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    """
    Default serializer used for user login. Please keep in mind that
    the authentication is done by separate function defined by
    :ref:`login-authenticator-setting` setting.

    By default :ref:`login-authenticator-setting` function will use
    :ref:`user-login-fields-setting` setting to extract the login field
    from the validated serializer data either by using the 'login' key
    (which is used here) or the specific login field name(s)
    (e.g. 'username', 'email').

    If you want different behavior, you need to
    override :ref:`login-authenticator-setting` in your settings.
    """
    login = serializers.CharField()
    password = serializers.CharField()


class DefaultRegisterEmailSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    """
    Default serializer used for e-mail registration (e-mail change).
    """
    email = serializers.EmailField(required=True)


class DefaultSendResetPasswordLinkSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    """
    Default serializer used for sending reset password link.

    It will use :ref:`send-reset-password-link-serializer-use-email-setting`
    setting.
    """

    def get_fields(self):
        fields = super().get_fields()
        if registration_settings.SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL:
            fields['email'] = serializers.CharField(required=True)
        else:
            fields['login'] = serializers.CharField(required=True)
        return fields


class DefaultUserProfileSerializer(serializers.ModelSerializer):
    """
    Default serializer used for user profile. It will use these:

    * User fields
    * :ref:`user-hidden-fields-setting` setting
    * :ref:`user-public-fields-setting` setting
    * :ref:`user-editable-fields-setting` setting

    to automagically generate the required serializer fields.
    """

    def __init__(self, *args, **kwargs):
        user_class = get_user_model()
        field_names = get_user_public_field_names()
        read_only_field_names = get_user_public_field_names(read_only=True)
        meta_obj = MetaObj()
        meta_obj.model = user_class
        meta_obj.fields = field_names
        meta_obj.read_only_fields = read_only_field_names
        self.Meta = meta_obj  # pylint: disable=invalid-name
        super().__init__(*args, **kwargs)


class DefaultRegisterUserSerializer(
        PasswordConfirmSerializerMixin,
        serializers.ModelSerializer):
    """
    Default serializer used for user registration. It will use these:

    * User fields
    * :ref:`user-hidden-fields-setting` setting
    * :ref:`user-public-fields-setting` setting

    to automagically generate the required serializer fields.
    """

    def __init__(self, *args, **kwargs):
        user_class = get_user_model()
        field_names = get_user_public_field_names(write_once=True)
        read_only_field_names = get_user_public_field_names(
            write_once=True, read_only=True)
        meta_obj = MetaObj()
        meta_obj.model = user_class
        meta_obj.fields = field_names
        meta_obj.read_only_fields = read_only_field_names
        self.Meta = meta_obj  # pylint: disable=invalid-name
        super().__init__(*args, **kwargs)

    def has_password_confirm_field(self):
        return registration_settings.REGISTER_SERIALIZER_PASSWORD_CONFIRM

    def validate(self, attrs):
        validators = [
            validate_user_password,
        ]
        if self.has_password_confirm_field():
            validators.append(validate_user_password_confirm)
        run_validators(validators, attrs)
        return attrs

    def create(self, validated_data):
        data = validated_data.copy()
        if self.has_password_confirm_field():
            del data['password_confirm']
        return self.Meta.model.objects.create_user(**data)



from functools import partial
from typing import Type

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.checks import register
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from rest_framework.settings import api_settings

from rest_registration.auth_token_managers import AbstractAuthTokenManager
from rest_registration.enums import ErrorCode, WarningCode
from rest_registration.notifications.email import parse_template_config
from rest_registration.settings import registration_settings
from rest_registration.utils.checks import no_exception_check, predicate_check
from rest_registration.utils.common import implies
from rest_registration.utils.users import (
    get_user_email_field_name,
    get_user_login_field_names,
    is_model_field_unique
)


@register()
@predicate_check(
    'django.contrib.auth is not in INSTALLED_APPS',
    ErrorCode.NO_AUTH_INSTALLED,
)
def auth_installed_check() -> bool:
    return _is_auth_installed()


@register()
@predicate_check(
    'RESET_PASSWORD_VERIFICATION_URL is not set',
    ErrorCode.NO_RESET_PASSWORD_VER_URL,
)
def reset_password_verification_url_check() -> bool:
    return implies(
        registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED,
        registration_settings.RESET_PASSWORD_VERIFICATION_URL,
    )


@register()
@predicate_check(
    'register verification is enabled,'
    ' but REGISTER_VERIFICATION_URL is not set',
    ErrorCode.NO_REGISTER_VER_URL,
)
def register_verification_url_check() -> bool:
    return implies(
        registration_settings.REGISTER_VERIFICATION_ENABLED,
        registration_settings.REGISTER_VERIFICATION_URL,
    )


@register()
@predicate_check(
    'register email verification is enabled,'
    ' but REGISTER_EMAIL_VERIFICATION_URL is not set',
    ErrorCode.NO_REGISTER_EMAIL_VER_URL,
)
def register_email_verification_url_check() -> bool:
    return implies(
        registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED,
        registration_settings.REGISTER_EMAIL_VERIFICATION_URL,
    )


@register()
@predicate_check(
    'VERIFICATION_FROM_EMAIL is not set',
    ErrorCode.NO_VER_FROM_EMAIL,
)
def verification_from_check() -> bool:
    return implies(
        any([
            registration_settings.REGISTER_VERIFICATION_ENABLED,
            registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED,
            registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED,
        ]),
        registration_settings.VERIFICATION_FROM_EMAIL,
    )


@register()
@predicate_check(
    'LOGIN_RETRIEVE_TOKEN is set but'
    ' the matching token authentication class is not'
    ' in DEFAULT_AUTHENTICATION_CLASSES',
    ErrorCode.NO_TOKEN_AUTH_CONFIG,
)
def token_auth_config_check() -> bool:
    return implies(
        _is_auth_installed() and registration_settings.LOGIN_RETRIEVE_TOKEN,
        _is_auth_token_manager_auth_class_enabled
    )


@register()
@predicate_check(
    'LOGIN_RETRIEVE_TOKEN is set but'
    ' the required Django apps are not in INSTALLED_APPS',
    ErrorCode.NO_TOKEN_AUTH_INSTALLED,
)
def token_auth_installed_check() -> bool:
    return implies(
        _is_auth_installed() and registration_settings.LOGIN_RETRIEVE_TOKEN,
        _is_auth_token_manager_app_name_installed,
    )


def _is_auth_token_manager_auth_class_enabled() -> bool:
    auth_token_manager = _get_auth_token_manager()
    auth_cls = auth_token_manager.get_authentication_class()
    return any(
        issubclass(cls, auth_cls)
        for cls in api_settings.DEFAULT_AUTHENTICATION_CLASSES
    )


def _is_auth_token_manager_app_name_installed() -> bool:
    auth_token_manager = _get_auth_token_manager()
    app_names = auth_token_manager.get_app_names()
    return all(app_name in settings.INSTALLED_APPS for app_name in app_names)


def _get_auth_token_manager() -> AbstractAuthTokenManager:
    auth_token_manager_cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
    auth_token_manager = auth_token_manager_cls()
    return auth_token_manager


@register()
@predicate_check(
    'REGISTER_VERIFICATION_AUTO_LOGIN is enabled,'
    ' but REGISTER_VERIFICATION_ONE_TIME_USE is not enabled.'
    ' This can allow multiple logins using the verification url.',
    WarningCode.REGISTER_VERIFICATION_MULTIPLE_AUTO_LOGIN,
)
def register_verification_one_time_auto_login_check() -> bool:
    return implies(
        all([
            registration_settings.REGISTER_VERIFICATION_ENABLED,
            registration_settings.REGISTER_VERIFICATION_AUTO_LOGIN,
        ]),
        registration_settings.REGISTER_VERIFICATION_ONE_TIME_USE,
    )


@register()
@no_exception_check(
    'REGISTER_VERIFICATION_EMAIL_TEMPLATES is invalid',
    ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
)
def valid_register_verification_email_template_config_check() -> None:
    _validate_email_template_config(
        registration_settings.REGISTER_VERIFICATION_ENABLED,
        registration_settings.REGISTER_VERIFICATION_EMAIL_TEMPLATES,
    )


@register()
@no_exception_check(
    'RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES is invalid',
    ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
)
def valid_reset_password_verification_email_template_config_check() -> None:
    _validate_email_template_config(
        registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED,
        registration_settings.RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES,
    )


@register()
@no_exception_check(
    'REGISTER_EMAIL_VERIFICATION_EMAIL_TEMPLATES is invalid',
    ErrorCode.INVALID_EMAIL_TEMPLATE_CONFIG,
)
def valid_register_email_verification_email_template_config_check() -> None:
    _validate_email_template_config(
        registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED,
        registration_settings.REGISTER_EMAIL_VERIFICATION_EMAIL_TEMPLATES,
    )


def _is_email_template_config_valid(template_config_data) -> bool:
    try:
        parse_template_config(template_config_data)
    except ImproperlyConfigured:
        return False
    else:
        return True


def _validate_email_template_config(enabled, template_config_data) -> None:
    if not enabled:
        return
    parse_template_config(template_config_data)


# The backends below allow inactive users to autenticate, which makes them
# incompatible with Django REST Framework authentication classes
# which check whether the is_active flag is set before authenticating.
#
# See https://github.com/apragacz/django-rest-registration/issues/52
#
__DRF_INCOMPATIBLE_DJANGO_AUTH_BACKENDS__ = [
    'django.contrib.auth.backends.AllowAllUsersModelBackend',
    'django.contrib.auth.backends.AllowAllUsersRemoteUserBackend',
]


@register()
@predicate_check(
    'One of following Django authentication backends (which are incompatible'
    ' with Django REST Framework authentication classes) is being used:\n\n'
    + '\n'.join(__DRF_INCOMPATIBLE_DJANGO_AUTH_BACKENDS__),
    ErrorCode.DRF_INCOMPATIBLE_DJANGO_AUTH_BACKEND,
)
def drf_compatible_django_auth_backend_check() -> bool:
    return all(
        incompat_cls_name not in settings.AUTHENTICATION_BACKENDS
        for incompat_cls_name in __DRF_INCOMPATIBLE_DJANGO_AUTH_BACKENDS__
    )


@register()
@predicate_check(
    'At least one of LOGIN_FIELDS is not unique',
    ErrorCode.LOGIN_FIELDS_NOT_UNIQUE,
)
def login_fields_unique_check() -> bool:
    return implies(_is_auth_installed(), _are_login_fields_unique)


def _are_login_fields_unique() -> bool:
    user_cls = get_user_model()  # type: Type[Model]
    user_meta = user_cls._meta  # pylint: disable=protected-access
    return all(
        is_model_field_unique(user_meta.get_field(field_name))
        for field_name in get_user_login_field_names())


@register()
@predicate_check(
    'AUTH_TOKEN_MANAGER_CLASS is not proper subclass'
    ' of AbstractAuthTokenManager',
    ErrorCode.INVALID_AUTH_TOKEN_MANAGER_CLASS,
)
def valid_auth_token_manager_class_proper_subclass_check() -> bool:
    return implies(
        _is_auth_installed(),
        _is_auth_token_manager_proper_subclass,
    )


@register()
@predicate_check(
    'AUTH_TOKEN_MANAGER_CLASS is not implementing'
    ' method get_authentication_class',
    ErrorCode.INVALID_AUTH_TOKEN_MANAGER_CLASS,
)
def valid_auth_token_manager_class_get_authentication_class_check() -> bool:
    return implies(
        _is_auth_installed(),
        partial(
            _is_auth_token_manager_class_implementing_method,
            'get_authentication_class'),
    )


@register()
@predicate_check(
    'AUTH_TOKEN_MANAGER_CLASS is not implementing method provide_token',
    ErrorCode.INVALID_AUTH_TOKEN_MANAGER_CLASS,
)
def valid_auth_token_manager_class_provide_token_check() -> bool:
    return implies(
        _is_auth_installed(),
        partial(
            _is_auth_token_manager_class_implementing_method,
            'provide_token'),
    )


# TODO: Issue #114 - remove deprecation check
@register()
@predicate_check(
    'LOGIN_SERIALIZER_CLASS contains deprecated get_authenticated_user method,'
    ' which will be removed in version 0.7.0;'
    ' for a replacement, please refer to LOGIN_AUTHENTICATOR setting',
    WarningCode.DEPRECATION
)
def deprecated_login_serializer_check() -> bool:
    serializer_class = registration_settings.LOGIN_SERIALIZER_CLASS
    deprecated = (
        hasattr(serializer_class, 'get_authenticated_user')
        and callable(serializer_class.get_authenticated_user))
    return not deprecated


# TODO: Issue #114 - remove deprecation check
@register()
@predicate_check(
    'SEND_RESET_PASSWORD_LINK_SERIALIZER_CLASS contains deprecated'
    ' get_user_or_none method, which will be removed in version 0.7.0;'
    ' for a replacement, please refer to SEND_RESET_PASSWORD_LINK_USER_FINDER setting',
    WarningCode.DEPRECATION
)
def deprecated_send_reset_password_link_serializer_check() -> bool:
    serializer_class = registration_settings.SEND_RESET_PASSWORD_LINK_SERIALIZER_CLASS
    deprecated = (
        hasattr(serializer_class, 'get_user_or_none')
        and callable(serializer_class.get_user_or_none))
    return not deprecated


# TODO: Issue #114 - remove deprecation check
@register()
@predicate_check(
    'REGISTER_EMAIL_SERIALIZER_CLASS contains deprecated get_email method,'
    ' which will be removed in version 0.7.0',
    WarningCode.DEPRECATION
)
def deprecated_register_email_serializer_check() -> bool:
    serializer_class = registration_settings.REGISTER_EMAIL_SERIALIZER_CLASS
    deprecated = (
        hasattr(serializer_class, 'get_email')
        and callable(serializer_class.get_email))
    return not deprecated


@register()
@predicate_check(
    'REGISTER_EMAIL_SERIALIZER_CLASS does not contain email field',
    ErrorCode.INVALID_REGISTER_EMAIL_SERIALIZER_CLASS
)
def valid_register_email_serializer_check() -> bool:
    serializer_class = registration_settings.REGISTER_EMAIL_SERIALIZER_CLASS
    serializer = serializer_class()
    try:
        return bool(serializer.fields['email'])
    except KeyError:
        return False


@register()
@predicate_check(
    'SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL is set'
    ' but email field is not unique',
    ErrorCode.NON_UNIQUE_FIELD_USED_AS_UNIQUE
)
def send_reset_password_link_serializer_email_unique_check() -> bool:
    return implies(
        registration_settings.SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL,
        _is_email_field_unique,
    )


def _is_email_field_unique() -> bool:
    user_cls = get_user_model()  # type: Type[Model]
    user_meta = user_cls._meta  # pylint: disable=protected-access
    email_field_name = get_user_email_field_name()
    return is_model_field_unique(user_meta.get_field(email_field_name))


def _is_auth_token_manager_proper_subclass() -> bool:
    cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
    return (
        issubclass(cls, AbstractAuthTokenManager)
        and cls != AbstractAuthTokenManager)


def _is_auth_token_manager_class_implementing_method(method_name: str) -> bool:
    cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
    cls_method = getattr(cls, method_name, None)
    abstract_method = getattr(AbstractAuthTokenManager, method_name)
    return cls_method is not abstract_method


def _is_auth_installed() -> bool:
    return 'django.contrib.auth' in settings.INSTALLED_APPS



from typing import Dict, List, Optional, Union

from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import APIException as _APIException
from rest_framework.settings import api_settings

from rest_registration.settings import registration_settings

DetailType = Union[str, List[str], Dict[str, List[str]]]


class APIException(_APIException):

    def __init__(
            self,
            detail: Optional[DetailType] = None,
            code: Optional[str] = None) -> None:
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code

        if registration_settings.USE_NON_FIELD_ERRORS_KEY_FROM_DRF_SETTINGS:
            detail = _wrap_detail_in_dict(detail)

        super().__init__(detail=detail, code=code)


class BadRequest(APIException):
    status_code = 400
    default_detail = _("Bad Request")
    default_code = 'bad-request'


class UserNotFound(BadRequest):
    default_detail = _("User not found")
    default_code = 'user-not-found'


class LoginInvalid(BadRequest):
    default_detail = _("Login or password invalid.")
    default_code = 'login-invalid'


class EmailAlreadyRegistered(BadRequest):
    default_detail = _("This email is already registered.")
    default_code = 'email-already-registered'


class UserWithoutEmailNonverifiable(BadRequest):
    default_detail = _("User without email cannot be verified")
    default_code = 'user-without-email-nonverifiable'


class AuthTokenError(BadRequest):
    default_detail = _("Could not process authentication token")
    default_code = 'auth-token-error'


class AuthTokenNotProvided(AuthTokenError):
    default_detail = _("Authentication token could not be provided")
    default_code = 'auth-token-not-provided'


class AuthTokenNotRevoked(AuthTokenError):
    default_detail = _("Authentication token cannot be revoked")
    default_code = 'auth-token-not-revoked'


class AuthTokenNotFound(AuthTokenError):
    default_detail = _("Authentication token not found")
    default_code = 'auth-token-not-found'


class SignatureError(BadRequest):
    default_detail = _("Generic signature error")
    default_code = 'signature-error'


class SignatureExpired(SignatureError):
    default_detail = _("Signature expired")
    default_code = 'signature-expired'


class SignatureInvalid(SignatureError):
    default_detail = _("Invalid signature")
    default_code = 'signature-invalid'


class VerificationTemplatesNotFound(APIException):
    status_code = 500
    default_detail = _("Could not find verification templates")
    default_code = 'verification-templates-not-found'


def _wrap_detail_in_dict(detail: DetailType) -> Dict[str, List[str]]:
    if isinstance(detail, list):
        return {api_settings.NON_FIELD_ERRORS_KEY: detail}
    if isinstance(detail, dict):
        return detail
    return {api_settings.NON_FIELD_ERRORS_KEY: [detail]}



from rest_registration.settings import registration_settings
from rest_registration.utils.signers import URLParamsSigner


class RegisterEmailSigner(URLParamsSigner):
    SALT_BASE = 'register-email'
    USE_TIMESTAMP = True

    def get_base_url(self):
        return registration_settings.REGISTER_EMAIL_VERIFICATION_URL

    def get_valid_period(self):
        return registration_settings.REGISTER_EMAIL_VERIFICATION_PERIOD



from enum import Enum


class NotificationType(Enum):
    REGISTER_VERIFICATION = 1
    REGISTER_EMAIL_VERIFICATION = 2
    RESET_PASSWORD_VERIFICATION = 3


class NotificationMethod(Enum):
    EMAIL = 1



from django.apps import AppConfig


class VerificationRedirectsConfig(AppConfig):
    name = 'rest_registration.contrib.verification_redirects'
    label = 'verification_redirects'
    verbose_name = 'Verification redirects'



from django.http import Http404
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.decorators import api_view, permission_classes

from rest_registration.api.serializers import PasswordConfirmSerializerMixin
from rest_registration.decorators import (
    api_view_serializer_class,
    api_view_serializer_class_getter
)
from rest_registration.exceptions import UserNotFound
from rest_registration.settings import registration_settings
from rest_registration.signers.reset_password import ResetPasswordSigner
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.users import get_user_by_verification_id
from rest_registration.utils.validation import (
    run_validators,
    validate_password_with_user_id,
    validate_user_password_confirm
)
from rest_registration.utils.verification import verify_signer_or_bad_request
from rest_registration.verification_notifications import (
    send_reset_password_verification_email_notification
)


@api_view_serializer_class_getter(
    lambda: registration_settings.SEND_RESET_PASSWORD_LINK_SERIALIZER_CLASS)
@api_view(['POST'])
@permission_classes(registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES)
def send_reset_password_link(request):
    '''
    Send email with reset password link.
    '''
    if not registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED:
        raise Http404()
    serializer_class = registration_settings.SEND_RESET_PASSWORD_LINK_SERIALIZER_CLASS
    serializer = serializer_class(data=request.data, context={'request': request})
    serializer.is_valid(raise_exception=True)
    if registration_settings.RESET_PASSWORD_FAIL_WHEN_USER_NOT_FOUND:
        success_message = _("Reset link sent")
    else:
        success_message = _("Reset link sent if the user exists in database")
    user_finder = registration_settings.SEND_RESET_PASSWORD_LINK_USER_FINDER
    try:
        user = user_finder(serializer.validated_data, serializer=serializer)
    except UserNotFound:
        if registration_settings.RESET_PASSWORD_FAIL_WHEN_USER_NOT_FOUND:
            raise
        return get_ok_response(success_message)
    send_reset_password_verification_email_notification(request, user)
    return get_ok_response(success_message)


class ResetPasswordSerializer(  # pylint: disable=abstract-method
        PasswordConfirmSerializerMixin,
        serializers.Serializer):
    user_id = serializers.CharField(required=True)
    timestamp = serializers.IntegerField(required=True)
    signature = serializers.CharField(required=True)
    password = serializers.CharField(required=True)

    def has_password_confirm_field(self):
        return registration_settings.RESET_PASSWORD_SERIALIZER_PASSWORD_CONFIRM

    def validate(self, attrs):
        validators = [
            validate_password_with_user_id,
        ]
        if self.has_password_confirm_field():
            validators.append(validate_user_password_confirm)
        run_validators(validators, attrs)
        return attrs


@api_view_serializer_class(ResetPasswordSerializer)
@api_view(['POST'])
@permission_classes(registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES)
def reset_password(request):
    '''
    Reset password, given the signature and timestamp from the link.
    '''
    process_reset_password_data(request.data, serializer_context={'request': request})
    return get_ok_response(_("Reset password successful"))


def process_reset_password_data(input_data, serializer_context=None):
    if serializer_context is None:
        serializer_context = {}
    if not registration_settings.RESET_PASSWORD_VERIFICATION_ENABLED:
        raise Http404()
    serializer = ResetPasswordSerializer(data=input_data, context=serializer_context)
    serializer.is_valid(raise_exception=True)

    data = serializer.validated_data.copy()
    password = data.pop('password')
    data.pop('password_confirm', None)
    signer = ResetPasswordSigner(data)
    verify_signer_or_bad_request(signer)

    user = get_user_by_verification_id(data['user_id'], require_verified=False)
    user.set_password(password)
    user.save()



from django.contrib.auth.password_validation import validate_password
from django.utils.translation import gettext as _
from rest_framework import permissions, serializers
from rest_framework.decorators import api_view, permission_classes

from rest_registration.api.serializers import PasswordConfirmSerializerMixin
from rest_registration.decorators import api_view_serializer_class
from rest_registration.settings import registration_settings
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.validation import validate_user_password_confirm


class ChangePasswordSerializer(  # pylint: disable=abstract-method
        PasswordConfirmSerializerMixin,
        serializers.Serializer):
    old_password = serializers.CharField()
    password = serializers.CharField()

    def has_password_confirm_field(self):
        return registration_settings.CHANGE_PASSWORD_SERIALIZER_PASSWORD_CONFIRM

    def validate_old_password(self, old_password):
        user = self.context['request'].user
        if not user.check_password(old_password):
            raise serializers.ValidationError(_("Old password is not correct"))
        return old_password

    def validate_password(self, password):
        user = self.context['request'].user
        validate_password(password, user=user)
        return password

    def validate(self, attrs):
        if self.has_password_confirm_field():
            validate_user_password_confirm(attrs)
        return attrs


@api_view_serializer_class(ChangePasswordSerializer)
@api_view(['POST'])
@permission_classes([permissions.IsAuthenticated])
def change_password(request):
    '''
    Change the user password.
    '''
    serializer = ChangePasswordSerializer(
        data=request.data,
        context={'request': request},
    )
    serializer.is_valid(raise_exception=True)

    user = request.user
    user.set_password(serializer.validated_data['password'])
    user.save()
    return get_ok_response(_("Password changed successfully"))



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Tuple,
    TypeVar,
    Union
)

from django.contrib import auth
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db.models.base import Model
from django.db.models.query import QuerySet
from django.http import Http404
from django.shortcuts import get_object_or_404 as _get_object_or_404

from rest_registration.exceptions import UserNotFound
from rest_registration.settings import registration_settings
from rest_registration.utils.common import set_or_none

try:
    from typing import Literal  # type: ignore
except ImportError:
    from typing_extensions import Literal


_DefaultT = TypeVar('_DefaultT')
_T = TypeVar('_T')
_ModelT = TypeVar('_ModelT', bound=Model)

if TYPE_CHECKING:
    from typing import Optional
    from django.contrib.auth.base_user import AbstractBaseUser
    from django.db.models import Field, ForeignObjectRel

    UserField = Union['Field', 'ForeignObjectRel']


_RAISE_EXCEPTION = object()


def authenticate_by_login_data(
        data: Dict[str, Any], **kwargs: Any) -> 'AbstractBaseUser':
    serializer = kwargs.get('serializer')
    if serializer:
        # TODO: Issue #114 - remove code supporting deprecated behavior
        get_authenticated_user = getattr(
            serializer, 'get_authenticated_user', None)
        if callable(get_authenticated_user):
            user = get_authenticated_user()  # type: Optional[AbstractBaseUser]
            if not user:
                raise UserNotFound()
            return user

    login_field_names = get_user_login_field_names()
    username_field_name = get_username_field_name()
    password = data.get('password')
    login_value = data.get('login')
    if password is None:
        raise UserNotFound()
    user_selectors = []  # type: List[Tuple[str, str]]
    if login_value is not None:
        user_selectors.extend(
            (field_name, login_value) for field_name in login_field_names)

    for field_name in login_field_names:
        field_value = data.get(field_name)
        if field_value is None:
            continue
        user_selectors.append((field_name, field_value))

    for field_name, field_value in user_selectors:
        if field_name == username_field_name:
            username = field_value
        else:
            user = get_user_by_lookup_dict(
                {field_name: field_value}, default=None, require_verified=False)
            if user is None:
                continue
            username = getattr(user, username_field_name)
        user = auth.authenticate(username=username, password=password)
        if user:
            return user

    raise UserNotFound()


def get_user_login_field_names() -> List[str]:
    user_class = get_user_model()
    return get_user_setting('LOGIN_FIELDS') or [user_class.USERNAME_FIELD]


def get_username_field_name() -> str:
    user_class = get_user_model()
    return user_class.USERNAME_FIELD


def get_user_verification_id(user: 'AbstractBaseUser') -> Any:
    verification_id_field = get_user_setting('VERIFICATION_ID_FIELD')
    return getattr(user, verification_id_field)


def get_user_by_verification_id(
        user_verification_id: Any,
        default: Union[_DefaultT, Literal[_RAISE_EXCEPTION]] = _RAISE_EXCEPTION,
        require_verified: bool = True) -> Union['AbstractBaseUser', _DefaultT]:
    verification_id_field = get_user_setting('VERIFICATION_ID_FIELD')
    return get_user_by_lookup_dict({
        verification_id_field: user_verification_id},
        default=default,
        require_verified=require_verified)


def find_user_by_by_send_reset_password_link_data(
        data: Dict[str, Any], **kwargs: Any) -> 'AbstractBaseUser':
    """
    Return user if matching given criteria (login fields / e-mail).
    Return ``None`` otherwise.
    """
    serializer = kwargs.get('serializer')
    if serializer:
        # TODO: Issue #114 - remove code supporting deprecated behavior
        get_user_or_none = getattr(serializer, 'get_user_or_none', None)
        if callable(get_user_or_none):
            user = get_user_or_none()  # type: Optional[AbstractBaseUser]
            if not user:
                raise UserNotFound()
            return user
    login_field_names = get_user_login_field_names()
    finder_tests = [('login', login_field_names)]
    finder_tests.extend((f, [f]) for f in login_field_names)

    for field_name, db_field_names in finder_tests:
        value = data.get(field_name)
        if value is None:
            continue
        for db_fn in db_field_names:
            user = get_user_by_lookup_dict(
                {db_fn: value}, default=None, require_verified=False)
            if user is not None:
                return user

    raise UserNotFound()


def user_with_email_exists(email: str) -> bool:
    user_class = get_user_model()
    email_field_name = get_user_email_field_name()
    if not email_field_name:
        return True
    queryset = user_class.objects.filter(**{email_field_name: email})
    return queryset.exists()


def is_user_email_field_unique() -> bool:
    email_field_name = get_user_email_field_name()
    email_field = get_user_field_obj(email_field_name)
    return is_model_field_unique(email_field)


def get_user_email_field_name() -> str:
    email = get_user_setting('EMAIL_FIELD')  # type: str
    return email


def get_user_by_lookup_dict(
        lookup_dict: Dict[str, Any],
        default: Union[_DefaultT, Literal[_RAISE_EXCEPTION]] = _RAISE_EXCEPTION,
        require_verified: bool = True) -> Union['AbstractBaseUser', _DefaultT]:
    verification_enabled = registration_settings.REGISTER_VERIFICATION_ENABLED
    verification_flag_field = get_user_setting('VERIFICATION_FLAG_FIELD')
    user_class = get_user_model()
    kwargs = {}
    kwargs.update(lookup_dict)
    if require_verified and verification_enabled and verification_flag_field:
        kwargs[verification_flag_field] = True
    try:
        queryset = user_class.objects.all()  # type: QuerySet[AbstractBaseUser]
        user = get_object_or_404(queryset, **kwargs)
    except Http404:
        if default is _RAISE_EXCEPTION:
            raise UserNotFound() from None
        return default
    else:
        return user


def get_user_field_obj(name: str) -> 'UserField':
    user_class = get_user_model()
    return user_class._meta.get_field(name)  # pylint: disable=protected-access


def get_user_setting(name: str) -> Any:
    setting_name = 'USER_{name}'.format(name=name)
    user_class = get_user_model()
    placeholder = object()
    value = getattr(user_class, name, placeholder)

    if value is placeholder:
        value = getattr(registration_settings, setting_name)

    return value


def build_initial_user(data: Dict[str, Any]) -> 'AbstractBaseUser':
    user_field_names = get_user_public_field_names(write_once=True)
    user_data = {}
    for field_name in user_field_names:
        if field_name in data:
            user_data[field_name] = data[field_name]
    user_class = get_user_model()
    return user_class(**user_data)


def get_user_field_names(
        allow_primary_key: bool = True,
        non_editable: bool = False) -> Tuple[str, ...]:

    def not_in_seq(names: Iterable[str]) -> Callable[[str], bool]:
        return lambda name: name not in names

    user_class = get_user_model()
    fields = user_class._meta.get_fields()  # pylint: disable=protected-access
    default_field_names = [f.name for f in fields
                           if (getattr(f, 'serialize', False) or
                               getattr(f, 'primary_key', False))]
    pk_field_names = [f.name for f in fields
                      if getattr(f, 'primary_key', False)]
    hidden_field_names = set(get_user_setting('HIDDEN_FIELDS'))
    hidden_field_names = hidden_field_names.union(['last_login', 'password'])
    public_field_names = get_user_setting('PUBLIC_FIELDS')
    editable_field_names = get_user_setting('EDITABLE_FIELDS')

    field_names = (public_field_names if public_field_names is not None
                   else default_field_names)
    if editable_field_names is None:
        editable_field_names = field_names

    editable_field_names = set(filter(not_in_seq(pk_field_names),
                                      editable_field_names))

    field_names = filter(not_in_seq(hidden_field_names), field_names)
    if not allow_primary_key:
        field_names = filter(not_in_seq(pk_field_names), field_names)

    if non_editable:
        field_names = filter(not_in_seq(editable_field_names), field_names)

    field_names_tuple = tuple(field_names)

    return field_names_tuple


def get_user_public_field_names(
        write_once: bool = False,
        read_only: bool = False) -> Tuple[str, ...]:

    fields = _get_user_default_fields()

    if read_only:
        field_names = _get_user_public_read_only_field_names(
            fields, write_once=write_once)
    else:
        field_names = _get_user_public_base_field_names(
            fields, write_once=write_once)

    return tuple(field_names)


def _get_user_public_read_only_field_names(
        fields: Iterable['UserField'],
        write_once: bool = False) -> List[str]:
    pk_name = _get_pk_name(fields)
    base_field_names = _get_user_public_base_field_names(
        fields, write_once=write_once)
    email_field_name = get_user_email_field_name()
    _editable_field_names = get_user_setting(
        'EDITABLE_FIELDS')  # type: Optional[Iterable[str]]
    editable_field_names = set_or_none(_editable_field_names)
    if write_once:
        excludes = {'last_login', pk_name}
    else:
        excludes = {'last_login', email_field_name, pk_name}
    field_names_rw = [f for f in base_field_names if f not in excludes]

    if not write_once and editable_field_names is not None:
        field_names_rw = [
            f for f in field_names_rw if f in editable_field_names]

    return [f for f in base_field_names if f not in field_names_rw]


def _get_user_public_base_field_names(
        fields: Iterable['UserField'],
        write_once: bool = False) -> List[str]:
    default_field_names = [f.name for f in fields]
    email_field_name = get_user_email_field_name()
    hidden_field_names = set(get_user_setting('HIDDEN_FIELDS'))
    _public_field_names = get_user_setting(
        'PUBLIC_FIELDS')  # type: Optional[Iterable[str]]
    public_field_names = set_or_none(_public_field_names)

    base_field_names = default_field_names
    if public_field_names is not None:
        allowed_field_names = public_field_names | {
            'password', email_field_name}
        base_field_names = [
            f for f in base_field_names if f in allowed_field_names]
    else:
        base_field_names = [
            f for f in base_field_names if f not in hidden_field_names]

    if not write_once:
        base_field_names = [f for f in base_field_names if f != 'password']

    return base_field_names


def _get_pk_name(fields: Iterable['UserField']) -> str:
    pk_names = [f.name for f in fields if getattr(f, 'primary_key', False)]
    if len(pk_names) != 1:
        raise ValueError('User model does not have one primary key')
    pk_name = pk_names[0]
    return pk_name


def _get_user_default_fields() -> List['UserField']:
    user_class = get_user_model()
    fields = user_class._meta.get_fields()  # pylint: disable=protected-access
    default_field_names = [
        f for f in fields
        if getattr(f, 'serialize', False) or getattr(f, 'primary_key', False)
    ]
    return default_field_names


def is_model_field_unique(field: 'UserField') -> bool:
    f_uniq = getattr(field, 'unique', False)  # type: bool
    f_pkey = getattr(field, 'primary_key', False)  # type: bool
    return f_uniq or f_pkey


def get_object_or_404(
        queryset: 'QuerySet[_ModelT]',
        *filter_args: Any,
        **filter_kwargs: Any) -> _ModelT:
    """
    Same as Django's standard shortcut, but make sure to also raise 404
    if the filter_kwargs don't match the required types.

    This function was copied from rest_framework.generics because of issue #36.
    """
    try:
        return _get_object_or_404(queryset, *filter_args, **filter_kwargs)
    except (TypeError, ValueError, ValidationError):
        raise Http404 from None



import datetime
from collections import OrderedDict, namedtuple
from textwrap import dedent

_Field = namedtuple('_Field', [
    'name',
    'default',
    'help',
    'import_string',
])


class Field(_Field):

    def __new__(
            cls, name, *,
            default=None,
            help=None,  # pylint: disable=redefined-builtin
            import_string=False):
        return super().__new__(
            cls, name=name, default=default,
            help=help, import_string=import_string)


USER_SETTINGS_FIELDS = [
    Field(
        'USER_LOGIN_FIELDS',
    ),
    Field(
        'USER_HIDDEN_FIELDS',
        default=(
            'last_login',
            'is_active',
            'is_staff',
            'is_superuser',
            'user_permissions',
            'groups',
            'date_joined',
        ),
    ),
    Field(
        'USER_PUBLIC_FIELDS',
    ),
    Field(
        'USER_EDITABLE_FIELDS',
    ),
    Field(
        'USER_EMAIL_FIELD',
        default='email',
    ),
    Field(
        'USER_VERIFICATION_ID_FIELD',
        default='pk',
        help=dedent("""\
            Field used in verification, as part of signed data.

            The given field should uniquely identify the user. This means that
            using any user field which could change over time
            (``email``, ``username``) is NOT recommended.
            """),
    ),
    Field(
        'USER_VERIFICATION_FLAG_FIELD',
        default='is_active',
    ),
]

REGISTER_SETTINGS_FIELDS = [
    Field(
        'REGISTER_SERIALIZER_CLASS',
        default='rest_registration.api.serializers.DefaultRegisterUserSerializer',  # noqa: E501,
        import_string=True,
        help=dedent("""\
            The serializer used by :ref:`register-view` endpoint.
            It is used to validate the input data and save (create)
            the newly registered user. You can use your custom serializer
            to customise validation logic and the way the user is created
            in the database.
            """),
    ),
    Field(
        'REGISTER_SERIALIZER_PASSWORD_CONFIRM',
        default=True,
        help=dedent("""\
            Used by ``DefaultRegisterUserSerializer``.
            If ``True``, the serializer requires
            additional field ``password_confirm`` which value should be
            the same as the value of ``password`` field.

            It may be useful to disable it if you perform password confirmation
            at the frontend level.
            """),
    ),

    Field(
        'REGISTER_OUTPUT_SERIALIZER_CLASS',
        default='rest_registration.api.serializers.DefaultUserProfileSerializer',  # noqa: E501,
        import_string=True,
        help=dedent("""\
            The default serializer used by register endpoint.
            It is used output the data associated with
            the newly registered user. You can use your custom serializer
            to customise the output representation of the user.
            """),
    ),

    Field(
        'REGISTER_VERIFICATION_ENABLED',
        default=True,
        help=dedent("""\
            If enabled, then newly registered user will not
            be verified (user field specified by
            :ref:`user-verification-flag-field-setting` will be false),
            and verification e-mail with activation link
            will be sent to the user email (specified by ``USER_EMAIL_FIELD``).
            """),
    ),
    Field(
        'REGISTER_VERIFICATION_PERIOD',
        default=datetime.timedelta(days=7),
        help=dedent("""\
            Specifies how long the activation link will be valid.
            """),
    ),
    Field(
        'REGISTER_VERIFICATION_URL',
        help=dedent("""\
            Frontend URL to which the query parameters will be appended
            to create the activation link for newly registered user.
            """),
    ),
    Field('REGISTER_VERIFICATION_ONE_TIME_USE', default=False),
    Field(
        'REGISTER_VERIFICATION_EMAIL_TEMPLATES',
        default={
            'subject': 'rest_registration/register/subject.txt',
            'body': 'rest_registration/register/body.txt',
        },
        help=dedent("""\
            Directory of templates used to generate the verification email.
            There are separate templates for email body and subject.
            If you want to generate html emails please refer to
            :ref:`html-email`.
            """),
    ),
    Field(
        'REGISTER_VERIFICATION_AUTO_LOGIN',
        default=False,
        help=dedent("""\
            Specifies whether a user will be logged in automatically when they
            verify their registration.
            """),
    ),
]

LOGIN_SETTINGS_FIELDS = [
    Field(
        'LOGIN_SERIALIZER_CLASS',
        default='rest_registration.api.serializers.DefaultLoginSerializer',  # noqa: E501
        import_string=True,
    ),
    Field(
        'LOGIN_AUTHENTICATOR',
        default='rest_registration.utils.users.authenticate_by_login_data',
        import_string=True,
        help=dedent("""\
            By default the login authenticator function will use
            :ref:`user-login-fields-setting` setting to extract the login field
            from the validated serializer data either by using the 'login' key
            or the specific login field name(s) (e.g. 'username', 'email').

            You can change that behavior by overriding this setting.

            The authenticator function receives these parameters as
            positional arguments:

            *   ``data`` - the validated data from the login serializer.

            and these parameters as keyword arguments:

            *   ``serializer`` - the source login serializer which generated
                the input data. This parameter could be dropped in the future,
                so it should be retrieved via ``kwargs.get()`` instead be named
                directly.

            If the user cannot be found, the function should raise ``UserNotFound``
            exception (from ``rest_registration.exceptions``).
            """),
    ),
    Field('LOGIN_AUTHENTICATE_SESSION'),
    Field('LOGIN_RETRIEVE_TOKEN'),
    Field(
        'AUTH_TOKEN_MANAGER_CLASS',
        default='rest_registration.auth_token_managers.RestFrameworkAuthTokenManager',  # noqa: E501
        import_string=True,
        help=dedent("""\
            The token manager class used by :ref:`login-view`
            and :ref:`logout-view` which provides an interface for providing
            and optionally revoking the token.
            The class should inherit from
            ``rest_registration.token_managers.AbstractTokenManager``.
            """)
    ),
]
RESET_PASSWORD_SETTINGS_FIELDS = [
    Field(
        'SEND_RESET_PASSWORD_LINK_SERIALIZER_CLASS',
        default='rest_registration.api.serializers.DefaultSendResetPasswordLinkSerializer',  # noqa: E501,
        import_string=True,
        help=dedent("""\
            The serializer used by :ref:`send-reset-password-link-view`
            endpoint. You can use your custom serializer
            to customise validation logic and perform additonal checks.
            Please remember that it should implement ``get_user_or_none``
            method which is used to obtain the user matching the criteria.
            """)
    ),
    Field(
        'SEND_RESET_PASSWORD_LINK_USER_FINDER',
        default='rest_registration.utils.users.find_user_by_by_send_reset_password_link_data',  # noqa: E501
        import_string=True,
        help=dedent("""\
            By default the user finder function will use
            :ref:`user-login-fields-setting` setting to extract the login field
            from the validated serializer data either by using the 'login' key
            or the specific login field name(s) (e.g. 'username', 'email').
            You can change that behavior by overriding this setting.

            The user finder function receives these parameters as
            positional arguments:

            *   ``data`` - the validated data from the send reset pasword
                link serializer.

            and these parameters as keyword arguments:

            *   ``serializer`` - the source send reset password link serializer
                which generated the input data. This parameter could be dropped
                in the future, so it should be retrieved via ``kwargs.get()``
                instead be named directly.

            If the user cannot be found, the function should raise ``UserNotFound``
            exception (from ``rest_registration.exceptions``).
            """)
    ),
    Field(
        'SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL',
        default=False,
        help=dedent("""\
            Used specifically by ``DefaultSendResetPasswordLinkSerializer``.

            If ``True``, use e-mail field instead of login fields to find
            the user who should receive the reset password link.
            """)
    ),
    Field(
        'RESET_PASSWORD_FAIL_WHEN_USER_NOT_FOUND',
        default=True,
        help=dedent("""\
            If ``True``, then reveal that the user does not exist
            while reset password link is being sent by signaling an error.
            """)
    ),
    Field('RESET_PASSWORD_VERIFICATION_ENABLED', default=True),
    Field(
        'RESET_PASSWORD_VERIFICATION_PERIOD',
        default=datetime.timedelta(days=1),
    ),
    Field('RESET_PASSWORD_VERIFICATION_URL'),
    Field('RESET_PASSWORD_VERIFICATION_ONE_TIME_USE', default=False),
    Field(
        'RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES',
        default={
            'subject': 'rest_registration/reset_password/subject.txt',
            'body': 'rest_registration/reset_password/body.txt',
        },
    ),
    Field(
        'RESET_PASSWORD_SERIALIZER_PASSWORD_CONFIRM',
        default=False,
        help=dedent("""\
            Used by ``ResetPasswordSerializer``.
            If ``True``, the serializer requires
            additional field ``password_confirm`` which value should be
            the same as the value of ``password`` field.

            It may be useful to disable it (this is currently the default)
            if you perform password confirmation
            at the frontend level.
            """),
    ),
]
REGISTER_EMAIL_SETTINGS_FIELDS = [
    Field(
        'REGISTER_EMAIL_SERIALIZER_CLASS',
        default='rest_registration.api.serializers.DefaultRegisterEmailSerializer',  # noqa: E501
        import_string=True,
        help=dedent("""\
            The serializer used by :ref:`register-email-view` endpoint.
            It is used to validate the input data and obtain new e-mail.
            You can use your custom serializer
            to customise validation logic. The important part is that it should contain
            ``email`` field.
            """),
    ),
    Field('REGISTER_EMAIL_VERIFICATION_ENABLED', default=True),
    Field(
        'REGISTER_EMAIL_VERIFICATION_PERIOD',
        default=datetime.timedelta(days=7),
    ),
    Field('REGISTER_EMAIL_VERIFICATION_URL'),
    Field(
        'REGISTER_EMAIL_VERIFICATION_EMAIL_TEMPLATES',
        default={
            'subject': 'rest_registration/register_email/subject.txt',
            'body': 'rest_registration/register_email/body.txt',
        },
    ),
]

GLOBAL_VERIFICATION_SETTINGS_FIELDS = [
    Field('VERIFICATION_FROM_EMAIL'),
    Field('VERIFICATION_REPLY_TO_EMAIL'),
    Field(
        'VERIFICATION_EMAIL_HTML_TO_TEXT_CONVERTER',
        default='rest_registration.utils.html.convert_html_to_text_preserving_urls',  # noqa: E501
        import_string=True,
    ),
    Field(
        'VERIFICATION_URL_BUILDER',
        default='rest_registration.utils.verification.build_default_verification_url',  # noqa: E501
        import_string=True,
        help=dedent("""\
            The builder function receives the ``signer`` object and construct
            the url using ``signer.get_base_url()``
            and ``signer.get_signed_data()``. The default url builder will use
            the base url and append the signed data as HTTP GET query string.
            It is be solely up to the implementer of custom builder function
            to encode the signed values properly in the URL.
            """),
    ),
    Field(
        'VERIFICATION_TEMPLATE_CONTEXT_BUILDER',
        default='rest_registration.utils.verification.build_default_template_context',  # noqa: E501
        import_string=True,
        help=dedent("""\
            The builder function receives these parameters as
            positional arguments:

            *   ``user`` - the user which is to be notified.
            *   ``user_address`` - the user address, which can be the
                user e-mail, phone number, etc.
            *   ``data`` - dictionary; in most cases it contains only
                one entry, which is the ``param_signer``
                under ``'param_signer'`` key. The implementer of the
                custom builder function should be aware that the contents
                of the dictionary are dynamic and write defensive code
                to account that.

            and these parameters as keyword arguments:

            *   ``notification_type`` - value of
                ``rest_registration.notifications.enums.NotificationType``
                enum.
            *   ``notification_method`` - value of
                ``rest_registration.notifications.enums.NotificationMethod``
                enum. Currently there is only one choice which is
                ``NotificationMethod.EMAIL``.

            It is possible that in the future, additional keyword arguments
            may be added. Therefore the implementer
            of the custom builder function should take account of that,
            for instance by adding ``**kwargs`` in the signature
            of the function.
            """),
    ),
    Field(
        'VERIFICATION_TEMPLATES_SELECTOR',
        default='rest_registration.utils.verification.select_default_templates',
        import_string=True,
        help=dedent("""\
            By default, the verification templates selector function will use
            templates defined by these settings:

            *   :ref:`register-verification-email-templates-setting`
            *   :ref:`register-email-verification-email-templates-setting`
            *   :ref:`reset-password-verification-email-templates-setting`

            depending on the notification type (register,
            registere email, reset password).
            You can change that behavior by overriding this setting.

            The verification template selector function receives these parameters
            as keyword arguments:

            *   ``request`` - the request which is source of sending given verification.
            *   ``user`` - the user which is to be notified.
            *   ``notification_type`` - value of
                ``rest_registration.notifications.enums.NotificationType``
                enum.
            *   ``notification_method`` - value of
                ``rest_registration.notifications.enums.NotificationMethod``

            It is possible that in the future, additional keyword arguments
            may be added. Therefore the implementer
            of the custom selector function should take account of that,
            for instance by adding ``**kwargs`` in the signature
            of the function.

            The verification template selector should raise
            ``VerificationTemplatesNotFound`` exception when no matching templates
            were found. In that case, Django REST Registration will fall back
            to the default behavior (described above).
            """),
    ),
]

CHANGE_PASSWORD_SETTINGS_FIELDS = [
    Field('CHANGE_PASSWORD_SERIALIZER_PASSWORD_CONFIRM', default=True),
]

PROFILE_SETTINGS_FIELDS = [
    Field(
        'PROFILE_SERIALIZER_CLASS',
        default='rest_registration.api.serializers.DefaultUserProfileSerializer',  # noqa: E501
        import_string=True,
    ),
]

MISC_SETTINGS_FIELDS = [
    Field(
        'SUCCESS_RESPONSE_BUILDER',
        default='rest_registration.utils.responses.build_default_success_response',  # noqa: E501
        import_string=True,
    ),
    Field(
        'USE_NON_FIELD_ERRORS_KEY_FROM_DRF_SETTINGS',
        default=False,
        help=dedent("""\
            If ``True``, the base API exception class will wrap the detail
            string message (or message list) into a dictionary with a key
            defined by ``REST_FRAMEWORK['NON_FIELD_ERRORS_KEY']``.
            """),
    ),
]

PERMISSIONS_SETTINGS_FIELDS = [
    Field(
        'NOT_AUTHENTICATED_PERMISSION_CLASSES',
        default=['rest_framework.permissions.AllowAny'],
        import_string=True,
        help=dedent("""\
            This parameter establishes the permissions of the views that must
            be accessible without logging in.
            Basically replace ``AllowAny`` with the specified class.
            """)
    ),
]

SETTINGS_FIELDS_GROUPS_MAP = OrderedDict([
    ('user', USER_SETTINGS_FIELDS),
    ('register', REGISTER_SETTINGS_FIELDS),
    ('login', LOGIN_SETTINGS_FIELDS),
    ('reset_password', RESET_PASSWORD_SETTINGS_FIELDS),
    ('register_email', REGISTER_EMAIL_SETTINGS_FIELDS),
    ('global_verification', GLOBAL_VERIFICATION_SETTINGS_FIELDS),
    ('change_password', CHANGE_PASSWORD_SETTINGS_FIELDS),
    ('profile', PROFILE_SETTINGS_FIELDS),
    ('misc', MISC_SETTINGS_FIELDS),
    ('permissons', PERMISSIONS_SETTINGS_FIELDS),
])

SETTINGS_FIELDS_GROUPS = list(SETTINGS_FIELDS_GROUPS_MAP.values())

SETTINGS_FIELDS = [f for fields in SETTINGS_FIELDS_GROUPS for f in fields]



from typing import Any, Dict, Optional

from rest_framework.response import Response

from rest_registration.settings import registration_settings


def get_ok_response(
        message: str,
        status: int = 200,
        extra_data: Optional[Dict[str, Any]] = None) -> Response:
    builder = registration_settings.SUCCESS_RESPONSE_BUILDER
    response = builder(
        message=message, status=status, extra_data=extra_data)  # type: Response
    return response


def build_default_success_response(
        message: str,
        status: int,
        extra_data: Optional[Dict[str, Any]]) -> Response:
    data = {'detail': message}
    if extra_data:
        data.update(extra_data)
    return Response(data, status=status)



"""
Custom signals sent during the registration and activation processes.
"""

from django.dispatch import Signal

# A new user has registered.
user_registered = Signal(providing_args=["user", "request"])

# A user has activated his or her account.
user_activated = Signal(providing_args=["user", "request"])

# A user has verified his or her new email.
user_changed_email = Signal(
    providing_args=["user", "new_email", "old_email", "request"]
)



from typing import TYPE_CHECKING, Any, Dict, Optional
from urllib.parse import urlencode

from django.core import signing
from rest_framework.request import Request

from rest_registration.exceptions import SignatureExpired, SignatureInvalid
from rest_registration.notifications.enums import NotificationMethod, NotificationType
from rest_registration.settings import registration_settings
from rest_registration.utils.signers import URLParamsSigner

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser


def verify_signer_or_bad_request(signer: URLParamsSigner) -> None:
    try:
        signer.verify()
    except signing.SignatureExpired:
        raise SignatureExpired() from None
    except signing.BadSignature:
        raise SignatureInvalid() from None


def build_default_verification_url(signer: URLParamsSigner) -> str:
    base_url = signer.get_base_url()
    params = urlencode(signer.get_signed_data())
    url = '{base_url}?{params}'.format(base_url=base_url, params=params)
    if signer.request:
        url = signer.request.build_absolute_uri(url)
    return url


def build_default_template_context(
        user: 'AbstractBaseUser',
        user_address: Any,
        data: Dict[str, Any],
        notification_type: Optional[NotificationType] = None,
        notification_method: Optional[NotificationMethod] = None) -> Dict[str, Any]:
    context = {
        'user': user,
        'email': user_address,
    }
    data = data.copy()
    params_signer = data.pop('params_signer', None)
    if params_signer:
        context['verification_url'] = params_signer.get_url()
    context.update(data)
    return context


def select_default_templates(
        request: Request,
        user: 'AbstractBaseUser',
        notification_type: NotificationType,
        notification_method: NotificationMethod) -> Dict[str, str]:
    mapping = _get_default_email_templates_mapping()
    assert notification_type in mapping
    # Note: in case of KeyError we should raise VerificationTemplatesNotFound,
    # but that's not expected as we're covering all the cases here.
    return mapping[notification_type]


def _get_default_email_templates_mapping() -> Dict[NotificationType, Dict[str, str]]:
    return {
        NotificationType.REGISTER_VERIFICATION: registration_settings.REGISTER_VERIFICATION_EMAIL_TEMPLATES,  # noqa: E501
        NotificationType.REGISTER_EMAIL_VERIFICATION: registration_settings.REGISTER_EMAIL_VERIFICATION_EMAIL_TEMPLATES,  # noqa: E501
        NotificationType.RESET_PASSWORD_VERIFICATION: registration_settings.RESET_PASSWORD_VERIFICATION_EMAIL_TEMPLATES,  # noqa: E501
    }
